#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
import six

from zun.common import exception
import zun.conf
from zun.db import api as dbapi
from zun.tests.unit.db import base
from zun.tests.unit.db import utils

CONF = zun.conf.CONF


class DbVolumeMappingTestCase(base.DbTestCase):

    def setUp(self):
        super(DbVolumeMappingTestCase, self).setUp()

    def test_create_volume_mapping(self):
        utils.create_test_volume_mapping(context=self.context)

    def test_create_volume_mapping_already_exists(self):
        utils.create_test_volume_mapping(context=self.context,
                                         uuid='123')
        with self.assertRaisesRegex(exception.VolumeMappingAlreadyExists,
                                    'A volume mapping with UUID 123.*'):
            utils.create_test_volume_mapping(context=self.context,
                                             uuid='123')

    def test_get_volume_mapping_by_uuid(self):
        volume_mapping = utils.create_test_volume_mapping(context=self.context)
        res = dbapi.get_volume_mapping_by_uuid(self.context,
                                               volume_mapping.uuid)
        self.assertEqual(volume_mapping.id, res.id)
        self.assertEqual(volume_mapping.uuid, res.uuid)

    def test_get_volume_mapping_that_does_not_exist(self):
        self.assertRaises(exception.VolumeMappingNotFound,
                          dbapi.get_volume_mapping_by_uuid,
                          self.context,
                          uuidutils.generate_uuid())

    def test_list_volume_mappings(self):
        uuids = []
        for i in range(1, 6):
            volume = utils.create_test_volume(
                uuid=uuidutils.generate_uuid(),
                context=self.context)
            volume_mapping = utils.create_test_volume_mapping(
                uuid=uuidutils.generate_uuid(),
                volume_id=volume.id,
                context=self.context)
            uuids.append(six.text_type(volume_mapping['uuid']))
        res = dbapi.list_volume_mappings(self.context)
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), sorted(res_uuids))

    def test_list_volume_mappings_sorted(self):
        uuids = []
        for i in range(5):
            volume = utils.create_test_volume(
                uuid=uuidutils.generate_uuid(),
                context=self.context)
            volume_mapping = utils.create_test_volume_mapping(
                uuid=uuidutils.generate_uuid(),
                volume_id=volume.id,
                context=self.context)
            uuids.append(six.text_type(volume_mapping.uuid))
        res = dbapi.list_volume_mappings(self.context, sort_key='uuid')
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), res_uuids)

        self.assertRaises(exception.InvalidParameterValue,
                          dbapi.list_volume_mappings,
                          self.context,
                          sort_key='foo')

    def test_list_volume_mappings_with_filters(self):
        volume1 = utils.create_test_volume(
            id=1,
            volume_provider='provider-one',
            uuid=uuidutils.generate_uuid(),
            context=self.context)
        volume_mapping1 = utils.create_test_volume_mapping(
            id=1,
            volume_id=volume1.id,
            uuid=uuidutils.generate_uuid(),
            context=self.context)
        volume2 = utils.create_test_volume(
            id=2,
            volume_provider='provider-two',
            uuid=uuidutils.generate_uuid(),
            context=self.context)
        volume_mapping2 = utils.create_test_volume_mapping(
            id=2,
            uuid=uuidutils.generate_uuid(),
            volume_id=volume2.id,
            context=self.context)

        res = dbapi.list_volume_mappings(
            self.context, filters={'volume_provider': 'provider-one'})
        self.assertEqual([volume_mapping1.id], [r.id for r in res])

        res = dbapi.list_volume_mappings(
            self.context, filters={'volume_provider': 'provider-two'})
        self.assertEqual([volume_mapping2.id], [r.id for r in res])

        res = dbapi.list_volume_mappings(
            self.context, filters={'volume_provider': 'bad-provider'})
        self.assertEqual([], [r.id for r in res])

        res = dbapi.list_volume_mappings(
            self.context,
            filters={'volume_provider': volume1.volume_provider})
        self.assertEqual([volume_mapping1.id], [r.id for r in res])

    def test_destroy_volume_mapping(self):
        volume_mapping = utils.create_test_volume_mapping(context=self.context)
        dbapi.destroy_volume_mapping(self.context, volume_mapping.id)
        self.assertRaises(exception.VolumeMappingNotFound,
                          dbapi.get_volume_mapping_by_uuid,
                          self.context, volume_mapping.uuid)

    def test_destroy_volume_mapping_by_uuid(self):
        volume_mapping = utils.create_test_volume_mapping(context=self.context)
        dbapi.destroy_volume_mapping(self.context, volume_mapping.uuid)
        self.assertRaises(exception.VolumeMappingNotFound,
                          dbapi.get_volume_mapping_by_uuid,
                          self.context, volume_mapping.uuid)

    def test_destroy_volume_mapping_that_does_not_exist(self):
        self.assertRaises(exception.VolumeMappingNotFound,
                          dbapi.destroy_volume_mapping, self.context,
                          uuidutils.generate_uuid())

    def test_update_volume_mapping(self):
        volume_mapping = utils.create_test_volume_mapping(context=self.context)
        old_path = volume_mapping.container_path
        new_path = 'new-container-path'
        self.assertNotEqual(old_path, new_path)

        res = dbapi.update_volume_mapping(self.context, volume_mapping.id,
                                          {'container_path': new_path})
        self.assertEqual(new_path, res.container_path)

    def test_update_volume_mapping_not_found(self):
        volume_mapping_uuid = uuidutils.generate_uuid()
        new_path = 'new-container-path'
        self.assertRaises(exception.VolumeMappingNotFound,
                          dbapi.update_volume_mapping, self.context,
                          volume_mapping_uuid,
                          {'container_path': new_path})

    def test_update_volume_mapping_uuid(self):
        volume_mapping = utils.create_test_volume_mapping(context=self.context)
        self.assertRaises(exception.InvalidParameterValue,
                          dbapi.update_volume_mapping, self.context,
                          volume_mapping.id, {'uuid': ''})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ipaddress
import math
import six
import sys
import time

from neutronclient.common import exceptions
from oslo_log import log as logging
from oslo_utils import excutils

from zun.common import exception
from zun.common.i18n import _
import zun.conf
from zun.network import network
from zun.network import neutron
from zun import objects
from zun.objects import fields as obj_fields
from zun.pci import manager as pci_manager
from zun.pci import utils as pci_utils
from zun.pci import whitelist as pci_whitelist

CONF = zun.conf.CONF

LOG = logging.getLogger(__name__)

BINDING_PROFILE = 'binding:profile'
BINDING_HOST_ID = 'binding:host_id'
DEVICE_OWNER = 'compute:kuryr'


class KuryrNetwork(network.Network):
    def init(self, context, docker_api):
        self.docker = docker_api
        self.neutron_api = neutron.NeutronAPI(context)
        self.context = context
        self.pci_whitelist = pci_whitelist.Whitelist(
            CONF.pci.passthrough_whitelist)
        self.last_neutron_extension_sync = None
        self.extensions = {}

    def create_network(self, name, neutron_net_id):
        """Create a docker network with Kuryr driver.

        The docker network to be created will be based on the specified
        neutron net. It is assumed that the neutron net will have one
        or two subnets. If there are two subnets, it must be a ipv4
        subnet and a ipv6 subnet and containers created from this network
        will have both ipv4 and ipv6 addresses.

        What this method does is finding the subnets under the specified
        neutron net, retrieving the cidr, gateway of each
        subnet, and compile the list of parameters for docker.create_network.
        """
        # find a v4 and/or v6 subnet of the network
        shared = \
            self.neutron_api.get_neutron_network(neutron_net_id)[
                'shared']
        subnets = self.neutron_api.list_subnets(network_id=neutron_net_id)
        subnets = subnets.get('subnets', [])
        v4_subnet = self._get_subnet(subnets, ip_version=4)
        v6_subnet = self._get_subnet(subnets, ip_version=6)
        if not v4_subnet and not v6_subnet:
            raise exception.ZunException(_(
                "The Neutron network %s has no subnet") % neutron_net_id)

        # IPAM driver specific options
        ipam_options = {
            "Driver": CONF.network.driver_name,
            "Options": {
                'neutron.net.shared': str(shared)
            },
            "Config": []
        }

        # Driver specific options
        options = {
            'neutron.net.uuid': neutron_net_id,
            'neutron.net.shared': str(shared)
        }

        if v4_subnet:
            ipam_options['Options']['neutron.subnet.uuid'] = \
                v4_subnet.get('id')
            ipam_options["Config"].append({
                "Subnet": v4_subnet['cidr'],
                "Gateway": v4_subnet['gateway_ip']
            })

            options['neutron.subnet.uuid'] = v4_subnet.get('id')
        if v6_subnet:
            ipam_options['Options']['neutron.subnet.v6.uuid'] = \
                v6_subnet.get('id')
            ipam_options["Config"].append({
                "Subnet": v6_subnet['cidr'],
                "Gateway": v6_subnet['gateway_ip']
            })

            options['neutron.subnet.v6.uuid'] = v6_subnet.get('id')

        network_dict = {}
        network_dict['project_id'] = self.context.project_id
        network_dict['user_id'] = self.context.user_id
        network_dict['name'] = name
        network_dict['neutron_net_id'] = neutron_net_id
        network = objects.Network(self.context, **network_dict)

        for attempt in (1, 2, 3):
            LOG.debug("Attempt (%s) to create network: %s", attempt, network)
            created_network = self._create_network_attempt(
                network, options, ipam_options)
            if created_network:
                return created_network
            backoff = int(math.pow(2, attempt) - 1)
            time.sleep(backoff)

        raise exception.ZunException(_(
            "Cannot create docker network after several attempts %s"))

    def _create_network_attempt(self, network, options, ipam_options):
        # The DB model has unique constraint on 'neutron_net_id' field
        # which will guarantee only one request can create the network in here
        # (and call docker.create_network later) if there are concurrent
        # requests on creating networks for the same neutron net.
        try:
            network.create(self.context)
        except exception.NetworkAlreadyExists as e:
            if e.field != 'neutron_net_id':
                raise

            networks = objects.Network.list(
                self.context,
                filters={'neutron_net_id': network.neutron_net_id})
            LOG.debug("network objects with 'neutron_net_id' as '%(net_id)s': "
                      "%(networks)s",
                      {"net_id": network.neutron_net_id,
                       "networks": networks})
            docker_networks = self.list_networks(names=[network.name])
            LOG.debug("docker networks with name matching '%(name)s': "
                      "%(networks)s",
                      {"name": network.name,
                       "networks": docker_networks})
            if (networks and networks[0].network_id and
                    docker_networks and
                    networks[0].network_id == docker_networks[0]['Id']):
                LOG.debug("Network (%s) has already been created in docker",
                          network.name)
                return networks[0]
            else:
                # Probably, there are concurrent requests on creating the
                # network but the network is yet created in Docker.
                # We return False and let the caller retry.
                return False

        LOG.debug("Calling docker.create_network to create network %s, "
                  "ipam_options %s, options %s",
                  network.name, ipam_options, options)
        enable_ipv6 = bool(options.get('neutron.subnet.v6.uuid'))
        try:
            docker_network = self.docker.create_network(
                name=network.name,
                driver=CONF.network.driver_name,
                enable_ipv6=enable_ipv6,
                options=options,
                ipam=ipam_options)
        except Exception:
            with excutils.save_and_reraise_exception():
                network.destroy()

        network.network_id = docker_network['Id']
        network.save()
        return network

    def _get_subnet(self, subnets, ip_version):
        subnets = [s for s in subnets if s['ip_version'] == ip_version]
        if len(subnets) == 0:
            return None
        elif len(subnets) == 1:
            return subnets[0]
        else:
            raise exception.ZunException(_(
                "Multiple Neutron subnets exist with ip version %s") %
                ip_version)

    def remove_network(self, network):
        self.docker.remove_network(network.name)
        network.destroy()

    def inspect_network(self, network_name):
        return self.docker.inspect_network(network_name)

    def list_networks(self, **kwargs):
        return self.docker.networks(**kwargs)

    def create_or_update_port(self, container, network_name,
                              requested_network, security_groups=None,
                              set_binding_host=False):
        if requested_network.get('port'):
            neutron_port_id = requested_network.get('port')
            neutron_port = self.neutron_api.get_neutron_port(neutron_port_id)
            # update device_id in port
            port_req_body = {'port': {'device_id': container.uuid}}
            if set_binding_host:
                port_req_body['port']['device_owner'] = DEVICE_OWNER
                port_req_body['port'][BINDING_HOST_ID] = container.host
            self.neutron_api.update_port(neutron_port_id, port_req_body,
                                         admin=True)

            # If there is pci_request_id, it should be a sriov port.
            # populate pci related info.
            pci_request_id = requested_network.get('pci_request_id')
            if pci_request_id:
                self._populate_neutron_extension_values(container,
                                                        pci_request_id,
                                                        port_req_body)
                self._populate_pci_mac_address(container,
                                               pci_request_id,
                                               port_req_body)
                # NOTE(hongbin): Use admin context here because non-admin
                # context might not be able to update some attributes
                # (i.e. binding:profile).
                self.neutron_api.update_port(neutron_port_id, port_req_body,
                                             admin=True)
        else:
            network = self.inspect_network(network_name)
            neutron_net_id = network['Options']['neutron.net.uuid']
            port_dict = {
                'network_id': neutron_net_id,
                'tenant_id': self.context.project_id,
                'device_id': container.uuid,
            }
            if set_binding_host:
                port_dict['device_owner'] = DEVICE_OWNER
                port_dict[BINDING_HOST_ID] = container.host
            ip_addr = requested_network.get("fixed_ip")
            if ip_addr:
                port_dict['fixed_ips'] = [{'ip_address': ip_addr}]
            if security_groups is not None:
                port_dict['security_groups'] = security_groups
            neutron_port = self.neutron_api.create_port({'port': port_dict},
                                                        admin=True)
            neutron_port = neutron_port['port']

        preserve_on_delete = requested_network['preserve_on_delete']
        addresses = []
        for fixed_ip in neutron_port['fixed_ips']:
            ip_address = fixed_ip['ip_address']
            ip = ipaddress.ip_address(six.text_type(ip_address))
            if ip.version == 4:
                addresses.append({
                    'addr': ip_address,
                    'version': 4,
                    'port': neutron_port['id'],
                    'subnet_id': fixed_ip['subnet_id'],
                    'preserve_on_delete': preserve_on_delete
                })
            else:
                addresses.append({
                    'addr': ip_address,
                    'version': 6,
                    'port': neutron_port['id'],
                    'subnet_id': fixed_ip['subnet_id'],
                    'preserve_on_delete': preserve_on_delete
                })

        return addresses, neutron_port

    def connect_container_to_network(self, container, network_name,
                                     requested_network, security_groups=None):
        """Connect container to the network

        This method will create a neutron port, retrieve the ip address(es)
        of the port, and pass them to docker.connect_container_to_network.
        """
        container_id = container.container_id

        addresses, original_port = self.create_or_update_port(
            container, network_name, requested_network, security_groups)

        ipv4_address = None
        ipv6_address = None
        for address in addresses:
            if address['version'] == 4:
                ipv4_address = address['addr']
            if address['version'] == 6:
                ipv6_address = address['addr']

        kwargs = {}
        if ipv4_address:
            kwargs['ipv4_address'] = ipv4_address
        if ipv6_address:
            kwargs['ipv6_address'] = ipv6_address
        try:
            self.docker.connect_container_to_network(
                container_id, network_name, **kwargs)
        except exception.DockerError:
            with excutils.save_and_reraise_exception():
                self.do_port_cleanup(addresses, original_port)
        return addresses

    def do_port_cleanup(self, addresses, port):
        preserve_flag = addresses[0].get('preserve_on_delete')
        port_id = port.get('id')
        if preserve_flag:
            port_req_body = {'port': {'device_id': '', 'device_owner': ''}}
            port_req_body['port'][BINDING_HOST_ID] = None
            port_req_body['port']['mac_address'] = port.get('mac_address')
            port_req_body['port'][BINDING_PROFILE] = \
                port.get(BINDING_PROFILE, {})

            try:
                # Requires admin creds to set port bindings
                self.neutron_api.update_port(port_id, port_req_body,
                                             admin=True)
            except exception.PortNotFound:
                LOG.debug('Unable to unbind port %s as it no longer '
                          'exists.', port_id)
            except Exception:
                LOG.exception("Unable to clear device ID for port '%s'",
                              port_id)
        else:
            try:
                self.neutron_api.delete_port(port_id)
            except exception.PortNotFound:
                LOG.debug('Unable to delete port %s as it no longer '
                          'exists.', port_id)

    def disconnect_container_from_network(self, container, network_name,
                                          neutron_network_id=None):
        container_id = container.container_id

        addrs_list = []
        if container.addresses and neutron_network_id:
            addrs_list = container.addresses.get(neutron_network_id, [])

        self._disconnect_container_from_network(container_id, network_name,
                                                addrs_list)

    def _disconnect_container_from_network(self, container_id, network_name,
                                           addrs_list):
        neutron_ports = set()
        all_ports = set()
        for addr in addrs_list:
            all_ports.add(addr['port'])
            if not addr['preserve_on_delete']:
                port_id = addr['port']
                neutron_ports.add(port_id)

        try:
            if container_id:
                self.docker.disconnect_container_from_network(container_id,
                                                              network_name)
        finally:
            for port_id in all_ports:
                try:
                    if port_id in neutron_ports:
                        self.neutron_api.delete_port(port_id)
                    else:
                        self._unbind_port(port_id)
                except exceptions.PortNotFoundClient:
                    LOG.warning('Maybe your libnetwork distribution do not '
                                'have patch https://review.openstack.org/#/c/'
                                '441024/ or neutron tag extension does not '
                                'supported or not enabled.')

    def _unbind_port(self, port_id):
        port_req_body = {'port': {'device_id': '', 'device_owner': ''}}
        port_req_body['port'][BINDING_HOST_ID] = None
        try:
            port = self.neutron_api.get_neutron_port(port_id)
        except exception.PortNotFound:
            LOG.debug('Unable to show port %s as it no longer '
                      'exists.', port_id)
            return
        except Exception:
            # NOTE: In case we can't retrieve the binding:profile assume
            # that they are empty
            LOG.exception("Unable to get binding:profile for port '%s'",
                          port_id)
            port_profile = {}
        else:
            port_profile = port.get(BINDING_PROFILE, {})
        # NOTE: We're doing this to remove the binding information
        # for the physical device but don't want to overwrite the other
        # information in the binding profile.
        for profile_key in ('pci_vendor_info', 'pci_slot'):
            if profile_key in port_profile:
                del port_profile[profile_key]
        port_req_body['port'][BINDING_PROFILE] = port_profile

        try:
            # Requires admin creds to set port bindings
            self.neutron_api.update_port(port_id, port_req_body,
                                         admin=True)
        except exception.PortNotFound:
            LOG.debug('Unable to unbind port %s as it no longer '
                      'exists.', port_id)
        except Exception:
            LOG.exception("Unable to clear device ID for port '%s'",
                          port_id)

    def add_security_groups_to_ports(self, container, security_group_ids):
        port_ids = set()
        for addrs_list in container.addresses.values():
            for addr in addrs_list:
                port_id = addr['port']
                port_ids.add(port_id)

        search_opts = {'tenant_id': self.context.project_id}
        neutron_ports = self.neutron_api.list_ports(
            **search_opts).get('ports', [])
        neutron_ports = [p for p in neutron_ports if p['id'] in port_ids]
        for port in neutron_ports:
            if 'security_groups' not in port:
                port['security_groups'] = []
            port['security_groups'].extend(security_group_ids)
            updated_port = {'security_groups': port['security_groups']}
            try:
                LOG.info("Adding security group %(security_group_ids)s "
                         "to port %(port_id)s",
                         {'security_group_ids': security_group_ids,
                          'port_id': port['id']})
                self.neutron_api.update_port(port['id'],
                                             {'port': updated_port},
                                             admin=True)
            except exceptions.NeutronClientException as e:
                exc_info = sys.exc_info()
                if e.status_code == 400:
                    raise exception.SecurityGroupCannotBeApplied(
                        six.text_type(e))
                else:
                    six.reraise(*exc_info)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception("Neutron Error:")

    def remove_security_groups_from_ports(self, container, security_group_ids):
        port_ids = set()
        for addrs_list in container.addresses.values():
            for addr in addrs_list:
                port_id = addr['port']
                port_ids.add(port_id)

        search_opts = {'tenant_id': self.context.project_id}
        neutron_ports = self.neutron_api.list_ports(
            **search_opts).get('ports', [])
        neutron_ports = [p for p in neutron_ports if p['id'] in port_ids]
        for port in neutron_ports:
            port['security_groups'].remove(security_group_ids[0])
            updated_port = {'security_groups': port['security_groups']}
            try:
                LOG.info("Removing security group %(security_group_ids)s "
                         "from port %(port_id)s",
                         {'security_group_ids': security_group_ids,
                          'port_id': port['id']})
                self.neutron_api.update_port(port['id'],
                                             {'port': updated_port},
                                             admin=True)
            except exceptions.NeutronClientException as e:
                exc_info = sys.exc_info()
                if e.status_code == 400:
                    raise exception.SecurityGroupCannotBeRemoved(
                        six.text_type(e))
                else:
                    six.reraise(*exc_info)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception("Neutron Error:")

    def _refresh_neutron_extensions_cache(self):
        """Refresh the neutron extensions cache when necessary."""
        if (not self.last_neutron_extension_sync or
            ((time.time() - self.last_neutron_extension_sync)
             >= CONF.neutron.extension_sync_interval)):
            extensions_list = self.neutron_api.list_extensions()['extensions']
            self.last_neutron_extension_sync = time.time()
            self.extensions.clear()
            self.extensions = {ext['name']: ext for ext in extensions_list}

    def _has_port_binding_extension(self, refresh_cache=False):
        if refresh_cache:
            self._refresh_neutron_extensions_cache()
        return "Port Binding" in self.extensions

    def _populate_neutron_extension_values(self, container,
                                           pci_request_id,
                                           port_req_body):
        """Populate neutron extension values for the instance.

        If the extensions loaded contain QOS_QUEUE then pass the rxtx_factor.
        """
        self._refresh_neutron_extensions_cache()
        has_port_binding_extension = (
            self._has_port_binding_extension())
        if has_port_binding_extension:
            self._populate_neutron_binding_profile(container,
                                                   pci_request_id,
                                                   port_req_body)

    def _populate_neutron_binding_profile(self, container, pci_request_id,
                                          port_req_body):
        """Populate neutron binding:profile.

        Populate it with SR-IOV related information
        """
        if pci_request_id:
            pci_dev = pci_manager.get_container_pci_devs(
                container, pci_request_id).pop()
            profile = self._get_pci_device_profile(pci_dev)
            port_req_body['port'][BINDING_PROFILE] = profile

    def _populate_pci_mac_address(self, container, pci_request_id,
                                  port_req_body):
        """Add the updated MAC address value to the update_port request body.

        Currently this is done only for PF passthrough.
        """
        if pci_request_id is not None:
            pci_devs = pci_manager.get_container_pci_devs(
                container, pci_request_id)
            if len(pci_devs) != 1:
                # NOTE(ndipanov): We shouldn't ever get here since
                # InstancePCIRequest instances built from network requests
                # only ever index a single device, which needs to be
                # successfully claimed for this to be called as part of
                # allocate_networks method
                LOG.error("PCI request %(pci_request_id)s does not have a "
                          "unique device associated with it. Unable to "
                          "determine MAC address",
                          {'pci_request_id': pci_request_id},
                          container=container)
                return
            pci_dev = pci_devs[0]
            if pci_dev.dev_type == obj_fields.PciDeviceType.SRIOV_PF:
                try:
                    mac = pci_utils.get_mac_by_pci_address(pci_dev.address)
                except exception.PciDeviceNotFoundById as e:
                    LOG.error("Could not determine MAC address for %(addr)s, "
                              "error: %(e)s",
                              {"addr": pci_dev.address, "e": e},
                              container=container)
                else:
                    port_req_body['port']['mac_address'] = mac

    def _get_pci_device_profile(self, pci_dev):
        dev_spec = self.pci_whitelist.get_devspec(pci_dev)
        if dev_spec:
            return {'pci_vendor_info': "%s:%s" % (pci_dev.vendor_id,
                                                  pci_dev.product_id),
                    'pci_slot': pci_dev.address,
                    'physical_network':
                        dev_spec.get_tags().get('physical_network')}
        raise exception.PciDeviceNotFound(node_id=pci_dev.compute_node_uuid,
                                          address=pci_dev.address)



# Copyright (c) 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from keystoneauth1 import exceptions as ks_exc
from oslo_log import log as logging
from stevedore import driver

from zun.common import consts
from zun.common import exception
import zun.conf
from zun.scheduler.client import report
from zun.scheduler import request_filter
from zun.scheduler import utils


CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


class SchedulerClient(object):
    """Client library for placing calls to the scheduler."""

    def __init__(self):
        self.placement_client = report.SchedulerReportClient()
        scheduler_driver = CONF.scheduler.driver
        self.driver = driver.DriverManager(
            "zun.scheduler.driver",
            scheduler_driver,
            invoke_on_load=True).driver
        self.traits_ensured = False

    def select_destinations(self, context, containers, extra_specs):
        LOG.debug("Starting to schedule for containers: %s",
                  [c.uuid for c in containers])

        if not self.traits_ensured:
            self.placement_client._ensure_traits(context, consts.CUSTOM_TRAITS)
            self.traits_ensured = True

        alloc_reqs_by_rp_uuid, provider_summaries, allocation_request_version \
            = None, None, None
        request_filter.process_reqspec(context, extra_specs)
        resources = utils.resources_from_request_spec(
            context, containers[0], extra_specs)

        try:
            res = self.placement_client.get_allocation_candidates(context,
                                                                  resources)
            (alloc_reqs, provider_summaries, allocation_request_version) = res
        except (ks_exc.EndpointNotFound,
                ks_exc.MissingAuthPlugin,
                ks_exc.Unauthorized,
                ks_exc.DiscoveryFailure,
                ks_exc.ConnectFailure):
            # We have to handle the case that we failed to connect to the
            # Placement service.
            alloc_reqs, provider_summaries, allocation_request_version = (
                None, None, None)
        if not alloc_reqs:
            LOG.info("Got no allocation candidates from the Placement "
                     "API. This could be due to insufficient resources "
                     "or a temporary occurrence as compute nodes start "
                     "up.")
            raise exception.NoValidHost(reason="")
        else:
            # Build a dict of lists of allocation requests, keyed by
            # provider UUID, so that when we attempt to claim resources for
            # a host, we can grab an allocation request easily
            alloc_reqs_by_rp_uuid = collections.defaultdict(list)
            for ar in alloc_reqs:
                for rp_uuid in ar['allocations']:
                    alloc_reqs_by_rp_uuid[rp_uuid].append(ar)

        selections = self.driver.select_destinations(
            context, containers, extra_specs, alloc_reqs_by_rp_uuid,
            provider_summaries, allocation_request_version)
        return selections

    def update_resource(self, node):
        node.save()
        # TODO(Shunli): Update the inventory here



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add memory info to compute node

Revision ID: 53a8b515057e
Revises: eeac0d191f5a
Create Date: 2017-04-13 10:12:41.088202

"""

# revision identifiers, used by Alembic.
revision = '53a8b515057e'
down_revision = 'eeac0d191f5a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('compute_node',
                  sa.Column('mem_total', sa.Integer(), nullable=False))
    op.add_column('compute_node',
                  sa.Column('mem_free', sa.Integer(), nullable=False))
    op.add_column('compute_node',
                  sa.Column('mem_available', sa.Integer(), nullable=False))



#    Copyright 2015 Intel, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import textwrap

import mock
import pep8

from zun.hacking import checks
from zun.tests import base


class HackingTestCase(base.BaseTestCase):
    """Hacking test class.

    This class tests the hacking checks zun .hacking.checks by passing
    strings to the check methods like the pep8/flake8 parser would. The parser
    loops over each line in the file and then passes the parameters to the
    check method. The parameter names in the check method dictate what type of
    object is passed to the check method. The parameter types are::

        logical_line: A processed line with the following modifications:
            - Multi-line statements converted to a single line.
            - Stripped left and right.
            - Contents of strings replaced with "xxx" of same length.
            - Comments removed.
        physical_line: Raw line of text from the input file.
        lines: a list of the raw lines from the input file
        tokens: the tokens that contribute to this logical line
        line_number: line number in the input file
        total_lines: number of lines in the input file
        blank_lines: blank lines before this one
        indent_char: indentation character in this file (" " or "\t")
        indent_level: indentation (with tabs expanded to multiples of 8)
        previous_indent_level: indentation on previous line
        previous_logical: previous logical line
        filename: Path of the file being run through pep8

    When running a test on a check method the return will be False/None if
    there is no violation in the sample input. If there is an error a tuple is
    returned with a position in the line, and a message. So to check the result
    just assertTrue if the check is expected to fail and assertFalse if it
    should pass.
    """
    # We are patching pep8 so that only the check under test is actually
    # installed.

    @mock.patch('pep8._checks',
                {'physical_line': {}, 'logical_line': {}, 'tree': {}})
    def _run_check(self, code, checker, filename=None):
        pep8.register_check(checker)

        lines = textwrap.dedent(code).strip().splitlines(True)

        checker = pep8.Checker(filename=filename, lines=lines)
        checker.check_all()
        checker.report._deferred_print.sort()
        return checker.report._deferred_print

    def _assert_has_errors(self, code, checker, expected_errors=None,
                           filename=None):
        actual_errors = [e[:3] for e in
                         self._run_check(code, checker, filename)]
        self.assertEqual(expected_errors or [], actual_errors)

    def _assert_has_no_errors(self, code, checker, filename=None):
        self._assert_has_errors(code, checker, filename=filename)

    def test_assert_equal_in(self):
        errors = [(1, 0, "Z338")]
        check = checks.assert_equal_in

        code = "self.assertEqual(a in b, True)"
        self._assert_has_errors(code, check, errors)

        code = "self.assertEqual('str' in 'string', True)"
        self._assert_has_errors(code, check, errors)

        code = "self.assertEqual(any(a==1 for a in b), True)"
        self._assert_has_no_errors(code, check)

        code = "self.assertEqual(True, a in b)"
        self._assert_has_errors(code, check, errors)

        code = "self.assertEqual(True, 'str' in 'string')"
        self._assert_has_errors(code, check, errors)

        code = "self.assertEqual(True, any(a==1 for a in b))"
        self._assert_has_no_errors(code, check)

        code = "self.assertEqual(a in b, False)"
        self._assert_has_errors(code, check, errors)

        code = "self.assertEqual('str' in 'string', False)"
        self._assert_has_errors(code, check, errors)

        code = "self.assertEqual(any(a==1 for a in b), False)"
        self._assert_has_no_errors(code, check)

        code = "self.assertEqual(False, a in b)"
        self._assert_has_errors(code, check, errors)

        code = "self.assertEqual(False, 'str' in 'string')"
        self._assert_has_errors(code, check, errors)

        code = "self.assertEqual(False, any(a==1 for a in b))"
        self._assert_has_no_errors(code, check)

    def test_assert_equal_true_or_false(self):
        errors = [(1, 0, "Z323")]
        check = checks.assert_equal_true_or_false

        code = "self.assertEqual(True, A)"
        self._assert_has_errors(code, check, errors)

        code = "self.assertEqual(False, A)"
        self._assert_has_errors(code, check, errors)

        code = "self.assertTrue()"
        self._assert_has_no_errors(code, check)

        code = "self.assertFalse()"
        self._assert_has_no_errors(code, check)

    def test_no_mutable_default_args(self):
        errors = [(1, 0, "Z322")]
        check = checks.no_mutable_default_args

        code = "def get_info_from_bdm(virt_type, bdm, mapping=[])"
        self._assert_has_errors(code, check, errors)

        code = "defined = []"
        self._assert_has_no_errors(code, check)

        code = "defined, undefined = [], {}"
        self._assert_has_no_errors(code, check)

    def test_assert_is_not_none(self):
        errors = [(1, 0, "Z302")]
        check = checks.assert_equal_not_none

        code = "self.assertEqual(A is not None)"
        self._assert_has_errors(code, check, errors)

        code = "self.assertIsNone()"
        self._assert_has_no_errors(code, check)

    def test_assert_true_isinstance(self):
        errors = [(1, 0, "Z316")]
        check = checks.assert_true_isinstance

        code = "self.assertTrue(isinstance(e, exception.BuilAbortException))"
        self._assert_has_errors(code, check, errors)

        code = "self.assertTrue()"
        self._assert_has_no_errors(code, check)

    def test_no_xrange(self):
        errors = [(1, 0, "Z339")]
        check = checks.no_xrange

        code = "xrange(45)"
        self._assert_has_errors(code, check, errors)

        code = "range(45)"
        self._assert_has_no_errors(code, check)

    def test_use_timeunitls_utcow(self):
        errors = [(1, 0, "Z310")]
        check = checks.use_timeutils_utcnow

        code = "datetime.now"
        self._assert_has_errors(code, check, errors)

        code = "datetime.utcnow"
        self._assert_has_errors(code, check, errors)

        code = "datetime.aa"
        self._assert_has_no_errors(code, check)

        code = "aaa"
        self._assert_has_no_errors(code, check)

    def test_dict_constructor_with_list_copy(self):
        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "    dict([(i, connect_info[i])"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "    attrs = dict([(k, _from_json(v))"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "        type_names = dict((value, key) for key, value in"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "   dict((value, key) for key, value in"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "foo(param=dict((k, v) for k, v in bar.items()))"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            " dict([[i,i] for i in range(3)])"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "  dd = dict([i,i] for i in range(3))"))))

        self.assertEqual(0, len(list(checks.dict_constructor_with_list_copy(
            "        create_kwargs = dict(snapshot=snapshot,"))))

        self.assertEqual(0, len(list(checks.dict_constructor_with_list_copy(
            "      self._render_dict(xml, data_el, data.__dict__)"))))

    def test_no_log_warn(self):
        errors = [(1, 0, "Z352")]
        check = checks.no_log_warn
        code = """
                  LOG.warn("LOG.warn is deprecated")
               """
        self._assert_has_errors(code, check, errors)

        code = """
                  LOG.warning("LOG.warn is deprecated")
               """
        self._assert_has_no_errors(code, check)

    def test_no_log_translations(self):
        for log in checks.log_levels:
            bad = 'LOG.%s(_("Bad"))' % log
            self.assertEqual(1, len(list(checks.no_translate_logs(bad))))
            # Catch abuses when used with a variable and not a literal
            bad = 'LOG.%s(_(msg))' % log
            self.assertEqual(1, len(list(checks.no_translate_logs(bad))))



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
Base API for Database
"""

from oslo_db import api as db_api

from zun.common import profiler
import zun.conf

"""Add the database backend mapping here"""

CONF = zun.conf.CONF
_BACKEND_MAPPING = {'sqlalchemy': 'zun.db.sqlalchemy.api'}
IMPL = db_api.DBAPI.from_config(CONF,
                                backend_mapping=_BACKEND_MAPPING,
                                lazy=True)


@profiler.trace("db")
def _get_dbdriver_instance():
    """Return a DB API instance."""
    return IMPL


@profiler.trace("db")
def list_containers(context, container_type, filters=None, limit=None,
                    marker=None, sort_key=None, sort_dir=None):
    """List matching containers.

    Return a list of the specified columns for all containers that match
    the specified filters.

    :param context: The security context
    :param container_type: The container type
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of containers to return.
    :param marker: the last item of the previous page; we return the next
                   result set.
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_containers(
        context, container_type, filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def create_container(context, values):
    """Create a new container.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the container, and several dicts which are
                   passed
                   into the Drivers when managing this container. For
                   example:
                   ::

                    {
                     'uuid': uuidutils.generate_uuid(),
                     'name': 'example',
                     'type': 'virt'
                    }
    :returns: A container.
    """
    return _get_dbdriver_instance().create_container(context, values)


@profiler.trace("db")
def get_container_by_uuid(context, container_type, container_uuid):
    """Return a container.

    :param context: The security context
    :param container_type: The container type
    :param container_uuid: The uuid of a container.
    :returns: A container.
    """
    return _get_dbdriver_instance().get_container_by_uuid(
        context, container_type, container_uuid)


@profiler.trace("db")
def get_container_by_name(context, container_type, container_name):
    """Return a container.

    :param context: The security context
    :param container_type: The container type
    :param container_name: The name of a container.
    :returns: A container.
    """
    return _get_dbdriver_instance().get_container_by_name(
        context, container_type, container_name)


@profiler.trace("db")
def destroy_container(context, container_type, container_id):
    """Destroy a container and all associated interfaces.

    :param context: Request context
    :param container_type: The container type
    :param container_id: The id or uuid of a container.
    """
    return _get_dbdriver_instance().destroy_container(
        context, container_type, container_id)


@profiler.trace("db")
def update_container(context, container_type, container_id, values):
    """Update properties of a container.

    :param context: Request context
    :param container_type: The container type
    :param container_id: The id or uuid of a container.
    :param values: The properties to be updated
    :returns: A container.
    :raises: ContainerNotFound
    """
    return _get_dbdriver_instance().update_container(
        context, container_type, container_id, values)


@profiler.trace("db")
def list_volume_mappings(context, filters=None, limit=None, marker=None,
                         sort_key=None, sort_dir=None):
    """List matching volume mappings.

    Return a list of the specified columns for all volume mappings that match
    the specified filters.

    :param context: The security context
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of containers to return.
    :param marker: the last item of the previous page; we return the next
                   result set.
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_volume_mappings(
        context, filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def count_volume_mappings(context, **filters):
    """Count matching volume mappings.

    Return the count of all volume mappings that match
    the specified filters.

    :param context: The security context
    :param filters: Filters to apply.
    :returns: The count of volume mapping.
    """
    return _get_dbdriver_instance().count_volume_mappings(context, **filters)


@profiler.trace("db")
def create_volume_mapping(context, values):
    """Create a volume mapping.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the volume mapping.
    :returns: A volume mapping.
    """
    return _get_dbdriver_instance().create_volume_mapping(context, values)


@profiler.trace("db")
def get_volume_mapping_by_uuid(context, vm_uuid):
    """Return a volume mapping.

    :param context: The security context
    :param vm_uuid: The uuid of a volume mapping.
    :returns: A volume mapping.
    """
    return _get_dbdriver_instance().get_volume_mapping_by_uuid(
        context, vm_uuid)


@profiler.trace("db")
def destroy_volume_mapping(context, vm_id):
    """Destroy a volume mapping.

    :param context: Request context
    :param vm_id: The id or uuid of a volume mapping.
    """
    return _get_dbdriver_instance().destroy_volume_mapping(context, vm_id)


@profiler.trace("db")
def update_volume_mapping(context, vm_id, values):
    """Update properties of a volume mapping.

    :param context: Request context
    :param vm_id: The id or uuid of a volume mapping.
    :param values: The properties to be updated
    :returns: A volume mapping.
    :raises: VolumeMappingNotFound
    """
    return _get_dbdriver_instance().update_volume_mapping(
        context, vm_id, values)


@profiler.trace("db")
def create_volume(context, values):
    """Create a volume.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the volume.
    :returns: A volume.
    """
    return _get_dbdriver_instance().create_volume(context, values)


@profiler.trace("db")
def get_volume_by_id(context, vol_id):
    """Return a volume

    :param context: The security context
    :param vol_id: The id of a volume.
    :returns: A volume.
    """
    return _get_dbdriver_instance().get_volume_by_id(
        context, vol_id)


@profiler.trace("db")
def destroy_volume(context, vol_id):
    """Destroy a volume.

    :param context: Request context
    :param vm_id: The id or uuid of a volume.
    """
    return _get_dbdriver_instance().destroy_volume(context, vol_id)


@profiler.trace("db")
def update_volume(context, vol_id, values):
    """Update properties of a volume.

    :param context: Request context
    :param vm_id: The id or uuid of a volume.
    :param values: The properties to be updated
    :returns: A volume.
    :raises: VolumeNotFound
    """
    return _get_dbdriver_instance().update_volume(
        context, vol_id, values)


@profiler.trace("db")
def destroy_zun_service(host, binary):
    """Destroys a zun_service record.

    :param host: The host on which the service resides.
    :param binary: The binary file name of the service.
    :returns: A zun service record.
    """
    return _get_dbdriver_instance().destroy_zun_service(host, binary)


@profiler.trace("db")
def update_zun_service(host, binary, values):
    """Update properties of a zun_service.

    :param host: The host on which the service resides.
    :param binary: The binary file name of the service.
    :param values: The attributes to be updated.
    :returns: A zun service record.
    """
    return _get_dbdriver_instance().update_zun_service(host, binary, values)


@profiler.trace("db")
def get_zun_service(context, host, binary):
    """Return a zun_service record.

    :param context: The security context
    :param host: The host where the binary is located.
    :param binary: The name of the binary.
    :returns: A zun_service record.
    """
    return _get_dbdriver_instance().get_zun_service(host, binary)


@profiler.trace("db")
def create_zun_service(values):
    """Create a new zun_service record.

    :param values: A dict containing several items used to identify
                   and define the zun_service record.
    :returns: A zun_service record.
    """
    return _get_dbdriver_instance().create_zun_service(values)


@profiler.trace("db")
def list_zun_services(context, filters=None, limit=None,
                      marker=None, sort_key=None, sort_dir=None):
    """Get matching zun_service records.

    Return a list of the specified columns for all zun_services
    those match the specified filters.

    :param context: The security context
    :param filters: Filters disbaled services. Defaults to None.
    :param limit: Maximum number of zun_services to return.
    :param marker: the last item of the previous page; we return the next
                   result set.
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_zun_services(
        filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def list_zun_services_by_binary(context, binary):
    """List matching zun services.

    Return a list of the specified binary.

    :param context: The security context
    :param binary: The name of the binary.
    :returns: A list of tuples of the specified binary.
    """
    return _get_dbdriver_instance().list_zun_services_by_binary(binary)


@profiler.trace("db")
def destroy_image(context, image_uuid):
    """Destroy a image"""
    return _get_dbdriver_instance().destroy_image(context, image_uuid)


@profiler.trace("db")
def pull_image(context, values):
    """Create a new image.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the image, and several dicts which are
                   passed
                   into the Drivers when managing this image. For
                   example:
                   ::

                    {
                     'uuid': uuidutils.generate_uuid(),
                     'repo': 'hello-world',
                     'tag': 'latest'
                    }
    :returns: An image.
    """
    return _get_dbdriver_instance().pull_image(context, values)


@profiler.trace("db")
def update_image(image_id, values):
    """Update properties of an image.

    :param image_id: The id or uuid of an image.
    :param values: The properties to be updated.
    :returns: An Image.
    :raises: ImageNotFound
    """
    return _get_dbdriver_instance().update_image(image_id, values)


@profiler.trace("db")
def list_images(context, filters=None,
                limit=None, marker=None,
                sort_key=None, sort_dir=None):
    """Get matching images.

    Return a list of the specified columns for all images that
    match the specified filters.

    :param context: The security context
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of images to return.
    :param marker: the last item of the previous page; we
                    return the next
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_images(
        context, filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def get_image_by_id(context, image_id):
    """Return an image.

    :param context: The security context
    :param image_id: The id of an image.
    :returns: An image.
    """
    return _get_dbdriver_instance().get_image_by_id(context, image_id)


@profiler.trace("db")
def get_image_by_uuid(context, image_uuid):
    """Return an image.

    :param context: The security context
    :param image_uuid: The uuid of an image.
    :returns: An image.
    """
    return _get_dbdriver_instance().get_image_by_uuid(context, image_uuid)


@profiler.trace("db")
def list_resource_providers(context, filters=None, limit=None, marker=None,
                            sort_key=None, sort_dir=None):
    """Get matching resource providers.

    Return a list of the specified columns for all resource providers that
    match the specified filters.

    :param context: The security context
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of resource providers to return.
    :param marker: the last item of the previous page; we
                    return the next
    :param sort_key: Attribute by which results should be sorted.
                    (asc, desc)
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_resource_providers(
        context, filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def create_resource_provider(context, values):
    """Create a new resource provider.

    :param context: The security context
    :param values: A dict containing several items used to identify and
                   track the resource provider, and several dicts which are
                   passed into the Drivers when managing this resource
                   provider.
    :returns: A resource provider.
    """
    return _get_dbdriver_instance().create_resource_provider(context, values)


@profiler.trace("db")
def get_resource_provider(context, provider_ident):
    """Return a resource provider.

    :param context: The security context
    :param provider_ident: The uuid or name of a resource provider.
    :returns: A resource provider.
    """
    return _get_dbdriver_instance().get_resource_provider(
        context, provider_ident)


@profiler.trace("db")
def destroy_resource_provider(context, provider_id):
    """Destroy a resource provider and all associated interfaces.

    :param context: Request context
    :param provider_id: The id or uuid of a resource provider.
    """
    return _get_dbdriver_instance().destroy_resource_provider(
        context, provider_id)


@profiler.trace("db")
def update_resource_provider(context, provider_id, values):
    """Update properties of a resource provider.

    :param context: Request context
    :param provider_id: The id or uuid of a resource provider.
    :param values: The properties to be updated
    :returns: A resource provider.
    :raises: ResourceProviderNotFound
    """
    return _get_dbdriver_instance().update_resource_provider(
        context, provider_id, values)


@profiler.trace("db")
def list_resource_classes(context, limit=None, marker=None, sort_key=None,
                          sort_dir=None):
    """Get matching resource classes.

    Return a list of the specified columns for all resource classes.

    :param context: The security context
    :param limit: Maximum number of resource classes to return.
    :param marker: the last item of the previous page; we
                    return the next
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_resource_classes(
        context, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def create_resource_class(context, values):
    """Create a new resource class.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the resource class, and several dicts which are
                   passed into the Drivers when managing this resource class.
    :returns: A resource class.
    """
    return _get_dbdriver_instance().create_resource_class(context, values)


@profiler.trace("db")
def get_resource_class(context, resource_ident):
    """Return a resource class.

    :param context: The security context
    :param resource_ident: The uuid or name of a resource class.
    :returns: A resource class.
    """
    return _get_dbdriver_instance().get_resource_class(
        context, resource_ident)


@profiler.trace("db")
def destroy_resource_class(context, resource_uuid):
    """Destroy a resource class and all associated interfaces.

    :param context: Request context
    :param resource_uuid: The uuid of a resource class.
    """
    return _get_dbdriver_instance().destroy_resource_class(
        context, resource_uuid)


@profiler.trace("db")
def update_resource_class(context, resource_uuid, values):
    """Update properties of a resource class.

    :param context: Request context
    :param resource_uuid: The uuid of a resource class.
    :param values: The properties to be updated
    :returns: A resource class.
    :raises: ResourceClassNotFound
    """
    return _get_dbdriver_instance().update_resource_class(
        context, resource_uuid, values)


@profiler.trace("db")
def list_inventories(context, filters=None, limit=None, marker=None,
                     sort_key=None, sort_dir=None):
    """List matching inventories.

    Return a list of the specified columns for all inventories that match
    the specified filters.

    :param context: The security context
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of inventories to return.
    :param marker: the last item of the previous page; we return the next
                   result set.
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_inventories(
        context, filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def create_inventory(context, provider_id, values):
    """Create a new inventory.

    :param context: The security context
    :param provider_id: The id of a resource provider.
    :param values: A dict containing several items used to identify
                   and track the inventory, and several dicts which are
                   passed into the Drivers when managing this inventory.
    :returns: An inventory.
    """
    return _get_dbdriver_instance().create_inventory(
        context, provider_id, values)


@profiler.trace("db")
def get_inventory(context, inventory_ident):
    """Return a inventory.

    :param context: The security context
    :param inventory_ident: The id or name of an inventory.
    :returns: An inventory.
    """
    return _get_dbdriver_instance().get_inventory(
        context, inventory_ident)


@profiler.trace("db")
def destroy_inventory(context, inventory_id):
    """Destroy an inventory and all associated interfaces.

    :param context: Request context
    :param inventory_id: The id of a inventory.
    """
    return _get_dbdriver_instance().destroy_inventory(context, inventory_id)


@profiler.trace("db")
def update_inventory(context, inventory_id, values):
    """Update properties of an inventory.

    :param context: Request context
    :param inventory_id: The id of an inventory.
    :param values: The properties to be updated
    :returns: An inventory.
    :raises: InventoryNotFound
    """
    return _get_dbdriver_instance().update_inventory(
        context, inventory_id, values)


@profiler.trace("db")
def list_allocations(context, filters=None, limit=None, marker=None,
                     sort_key=None, sort_dir=None):
    """List matching allocations.

    Return a list of the specified columns for all allocations that match
    the specified filters.

    :param context: The security context
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of allocations to return.
    :param marker: the last item of the previous page; we return the next
                   result set.
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_allocations(
        context, filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def create_allocation(context, values):
    """Create a new allocation.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the allocation, and several dicts which are
                   passed into the Drivers when managing this allocation.
    :returns: An allocation.
    """
    return _get_dbdriver_instance().create_allocation(context, values)


@profiler.trace("db")
def get_allocation(context, allocation_id):
    """Return an allocation.

    :param context: The security context
    :param allocation_id: The id of an allocation.
    :returns: An allocation.
    """
    return _get_dbdriver_instance().get_allocation(context, allocation_id)


@profiler.trace("db")
def destroy_allocation(context, allocation_id):
    """Destroy an allocation and all associated interfaces.

    :param context: Request context
    :param allocation_id: The id of an allocation.
    """
    return _get_dbdriver_instance().destroy_allocation(context, allocation_id)


@profiler.trace("db")
def update_allocation(context, allocation_id, values):
    """Update properties of an allocation.

    :param context: Request context
    :param allocation_id: The id of an allocation.
    :param values: The properties to be updated
    :returns: An allocation.
    :raises: AllocationNotFound
    """
    return _get_dbdriver_instance().update_allocation(
        context, allocation_id, values)


@profiler.trace("db")
def list_compute_nodes(context, filters=None, limit=None, marker=None,
                       sort_key=None, sort_dir=None):
    """List matching compute nodes.

    Return a list of the specified columns for all compute nodes that match
    the specified filters.

    :param context: The security context
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of compute nodes to return.
    :param marker: the last item of the previous page; we return the next
                   result set.
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_compute_nodes(
        context, filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def create_compute_node(context, values):
    """Create a new compute node.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the compute node, and several dicts which are
                   passed into the Drivers when managing this compute node.
    :returns: A compute node.
    """
    return _get_dbdriver_instance().create_compute_node(context, values)


@profiler.trace("db")
def get_compute_node(context, node_uuid):
    """Return a compute node.

    :param context: The security context
    :param node_uuid: The uuid of a compute node.
    :returns: A compute node.
    """
    return _get_dbdriver_instance().get_compute_node(context, node_uuid)


@profiler.trace("db")
def get_compute_node_by_hostname(context, hostname):
    """Return a compute node.

    :param context: The security context
    :param hostname: The hostname of a compute node.
    :returns: A compute node.
    """
    return _get_dbdriver_instance().get_compute_node_by_hostname(
        context, hostname)


@profiler.trace("db")
def destroy_compute_node(context, node_uuid):
    """Destroy a compute node and all associated interfaces.

    :param context: Request context
    :param node_uuid: The uuid of a compute node.
    """
    return _get_dbdriver_instance().destroy_compute_node(context, node_uuid)


@profiler.trace("db")
def update_compute_node(context, node_uuid, values):
    """Update properties of a compute node.

    :param context: Request context
    :param node_uuid: The uuid of a compute node.
    :param values: The properties to be updated
    :returns: A compute node.
    :raises: ComputeNodeNotFound
    """
    return _get_dbdriver_instance().update_compute_node(
        context, node_uuid, values)


@profiler.trace("db")
def get_pci_device_by_addr(node_id, dev_addr):
    """Get PCI device by address."""
    return _get_dbdriver_instance().get_pci_device_by_addr(node_id, dev_addr)


@profiler.trace("db")
def get_pci_device_by_id(id):
    """Get PCI device by id."""
    return _get_dbdriver_instance().get_pci_device_by_id(id)


@profiler.trace("db")
def get_all_pci_device_by_node(node_id):
    """Get all PCI devices for one host."""
    return _get_dbdriver_instance().get_all_pci_device_by_node(node_id)


@profiler.trace("db")
def get_all_pci_device_by_container_uuid(container_uuid):
    """Get PCI devices allocated to container."""
    return _get_dbdriver_instance().get_all_pci_device_by_container_uuid(
        container_uuid)


@profiler.trace("db")
def get_all_pci_device_by_parent_addr(node_id, parent_addr):
    """Get all PCI devices by parent address."""
    return _get_dbdriver_instance().get_all_pci_device_by_parent_addr(
        node_id, parent_addr)


@profiler.trace("db")
def destroy_pci_device(node_id, address):
    """Delete a PCI device record."""
    return _get_dbdriver_instance().destroy_pci_device(node_id, address)


@profiler.trace("db")
def update_pci_device(node_id, address, value):
    """Update a pci device."""
    return _get_dbdriver_instance().update_pci_device(node_id, address, value)


@profiler.trace("db")
def action_start(context, values):
    """Start an action for an container."""
    return _get_dbdriver_instance().action_start(context, values)


@profiler.trace("db")
def action_finish(context, values):
    """Start an action for an container."""
    return _get_dbdriver_instance().action_finish(context, values)


@profiler.trace("db")
def actions_get(context, uuid):
    """Get all container actions for the provided container."""
    return _get_dbdriver_instance().actions_get(context, uuid)


@profiler.trace("db")
def action_get_by_request_id(context, uuid, request_id):
    """Get the action by request_id and given container."""
    return _get_dbdriver_instance().action_get_by_request_id(context, uuid,
                                                             request_id)


@profiler.trace("db")
def action_event_start(context, values):
    """Start an event on an container action."""
    return _get_dbdriver_instance().action_event_start(context, values)


@profiler.trace("db")
def action_event_finish(context, values):
    """Finish an event on an container action."""
    return _get_dbdriver_instance().action_event_finish(context, values)


@profiler.trace("db")
def action_events_get(context, action_id):
    """Get the events by action id."""
    return _get_dbdriver_instance().action_events_get(context, action_id)


@profiler.trace("db")
def quota_create(context, project_id, resource, limit):
    """Create a quota for the given project and resource"""
    return _get_dbdriver_instance().quota_create(context, project_id,
                                                 resource, limit)


@profiler.trace("db")
def quota_get(context, project_id, resource):
    """Retrieve a quota or raise if it does not exist"""
    return _get_dbdriver_instance().quota_get(context, project_id,
                                              resource)


@profiler.trace("db")
def quota_get_all_by_project(context, project_id):
    """Retrieve all quotas associated with a given project"""
    return _get_dbdriver_instance().quota_get_all_by_project(context,
                                                             project_id)


@profiler.trace("db")
def quota_update(context, project_id, resource, limit):
    """Update a quota or raise if it does not exist"""
    return _get_dbdriver_instance().quota_update(context, project_id,
                                                 resource, limit)


@profiler.trace("db")
def quota_destroy(context, project_id, resource):
    """Destroy resource quota associated with a given project"""
    return _get_dbdriver_instance().quota_destroy(context, project_id,
                                                  resource)


@profiler.trace("db")
def quota_destroy_all_by_project(context, project_id):
    """Destroy all resource associated with a given project."""
    return _get_dbdriver_instance().quota_destroy_all_by_project(context,
                                                                 project_id)


@profiler.trace("db")
def quota_class_create(context, class_name, resource, limit):
    """Create a quota class for the given name and resource"""
    return _get_dbdriver_instance().quota_class_create(context, class_name,
                                                       resource, limit)


@profiler.trace("db")
def quota_class_get(context, class_name, resource):
    """Retrieve a quota class or raise if it does not exist"""
    return _get_dbdriver_instance().quota_class_get(context, class_name,
                                                    resource)


@profiler.trace("db")
def quota_class_get_default(context):
    """Retrieve all default quota"""
    return _get_dbdriver_instance().quota_class_get_default(context)


@profiler.trace("db")
def quota_class_get_all_by_name(context, class_name):
    """Retrieve all quotas associated with a given quota class"""
    return _get_dbdriver_instance().quota_class_get_all_by_name(
        context, class_name)


@profiler.trace("db")
def quota_class_update(context, class_name, resource, limit):
    """Update a quota class or raise if it does not exist"""
    return _get_dbdriver_instance().quota_class_update(context, class_name,
                                                       resource, limit)


def quota_usage_get_all_by_project(context, project_id):
    """Retrieve all usage associated with a given resource."""
    return _get_dbdriver_instance().quota_usage_get_all_by_project(context,
                                                                   project_id)


@profiler.trace("db")
def create_network(context, values):
    """Create a new network.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the network, and several dicts which are
                   passed
                   into the Drivers when managing this network. For
                   example:
                   ::
                    {
                     'uuid': uuidutils.generate_uuid(),
                     'name': 'example',
                     'type': 'virt'
                    }
    :returns: A network.
    """
    return _get_dbdriver_instance().create_network(context, values)


@profiler.trace("db")
def get_network_by_uuid(context, network_uuid):
    """Return a network.

    :param context: The security context
    :param network_uuid: The uuid of a network.
    :returns: A network.
    """
    return _get_dbdriver_instance().get_network_by_uuid(context, network_uuid)


@profiler.trace("db")
def list_networks(context, filters=None, limit=None, marker=None,
                  sort_key=None, sort_dir=None):
    """List matching networks.

    Return a list of the specified columns for all networks that match
    the specified filters.

    :param context: The security context
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of networks to return.
    :param marker: the last item of the previous page; we return the next
                   result set.
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_networks(
        context, filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def update_network(context, uuid, values):
    """Update properties of a network.

    :param context: Request context
    :param uuid: The id or uuid of a network.
    :param values: The properties to be updated
    :returns: A network.
    """
    return _get_dbdriver_instance().update_network(
        context, uuid, values)


@profiler.trace("db")
def destroy_network(context, network_uuid):
    """Destroy a network.

    :param context: Request context
    :param network_uuid: The uuid of a network.
    """
    return _get_dbdriver_instance().destroy_network(context, network_uuid)


@profiler.trace("db")
def create_exec_instance(context, values):
    """Create a new exec instance.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the exec instance, and several dicts which are
                   passed into the Drivers when managing this exec instance.
    :returns: An exec instance.
    """
    return _get_dbdriver_instance().create_exec_instance(context, values)


@profiler.trace("db")
def list_exec_instances(context, filters=None, limit=None, marker=None,
                        sort_key=None, sort_dir=None):
    """List matching exec instances.

    Return a list exec instances that match the specified filters.

    :param context: The security context
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of containers to return.
    :param marker: the last item of the previous page; we return the next
                   result set.
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of exec instances.
    """
    return _get_dbdriver_instance().list_exec_instances(
        context, filters, limit, marker, sort_key, sort_dir)


@profiler.trace('db')
def count_usage(context, container_type, project_id, flag):
    return _get_dbdriver_instance().count_usage(context, container_type,
                                                project_id, flag)


@profiler.trace("db")
def create_registry(context, values):
    """Create a new registry.

    :param context: The security context
    :param values: A dict containing several items used to identify
                   and track the registry, and several dicts which are
                   passed
                   into the Drivers when managing this registry.
    :returns: A registry.
    """
    return _get_dbdriver_instance().create_registry(context, values)


@profiler.trace("db")
def get_registry_by_id(context, registry_id):
    """Return a registry.

    :param context: The security context
    :param registry_id: The id of a registry.
    :returns: A registry.
    """
    return _get_dbdriver_instance().get_registry_by_id(
        context, registry_id)


@profiler.trace("db")
def get_registry_by_uuid(context, registry_uuid):
    """Return a registry.

    :param context: The security context
    :param registry_uuid: The uuid of a registry.
    :returns: A registry.
    """
    return _get_dbdriver_instance().get_registry_by_uuid(
        context, registry_uuid)


@profiler.trace("db")
def get_registry_by_name(context, registry_name):
    """Return a registry.

    :param context: The security context
    :param registry_name: The name of a registry.
    :returns: A registry.
    """
    return _get_dbdriver_instance().get_registry_by_name(
        context, registry_name)


@profiler.trace("db")
def list_registries(context, filters=None, limit=None, marker=None,
                    sort_key=None, sort_dir=None):
    """List matching registries.

    Return a list of the specified columns for all registries that match
    the specified filters.

    :param context: The security context
    :param filters: Filters to apply. Defaults to None.
    :param limit: Maximum number of registries to return.
    :param marker: the last item of the previous page; we return the next
                   result set.
    :param sort_key: Attribute by which results should be sorted.
    :param sort_dir: Direction in which results should be sorted.
                     (asc, desc)
    :returns: A list of tuples of the specified columns.
    """
    return _get_dbdriver_instance().list_registries(
        context, filters, limit, marker, sort_key, sort_dir)


@profiler.trace("db")
def update_registry(context, uuid, values):
    """Update properties of a registry.

    :param context: Request context
    :param uuid: The id or uuid of a registry.
    :param values: The properties to be updated
    :returns: A registry.
    """
    return _get_dbdriver_instance().update_registry(
        context, uuid, values)


@profiler.trace("db")
def destroy_registry(context, registry_uuid):
    """Destroy a registry.

    :param context: Request context
    :param registry_uuid: The uuid of a registry.
    """
    return _get_dbdriver_instance().destroy_registry(context, registry_uuid)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from zun.common import exception
from zun import objects
from zun.scheduler import chance_scheduler
from zun.tests import base
from zun.tests.unit.db import utils


class ChanceSchedulerTestCase(base.TestCase):
    """Test case for Chance Scheduler."""

    driver_cls = chance_scheduler.ChanceScheduler

    @mock.patch.object(driver_cls, 'hosts_up')
    @mock.patch('random.choice')
    def test_select_destinations(self, mock_random_choice, mock_hosts_up):
        all_hosts = ['host1', 'host2', 'host3', 'host4']

        def _return_hosts(*args, **kwargs):
            return all_hosts

        mock_random_choice.side_effect = ['host3']
        mock_hosts_up.side_effect = _return_hosts

        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        extra_spec = {}
        dests = self.driver_cls().select_destinations(self.context, containers,
                                                      extra_spec)

        self.assertEqual(1, len(dests))
        (host, node) = (dests[0]['host'], dests[0]['nodename'])
        self.assertEqual('host3', host)
        self.assertIsNone(node)

        calls = [mock.call(all_hosts)]
        self.assertEqual(calls, mock_random_choice.call_args_list)

    @mock.patch.object(driver_cls, 'hosts_up')
    def test_select_destinations_no_valid_host(self, mock_hosts_up):

        def _return_no_host(*args, **kwargs):
            return []

        mock_hosts_up.side_effect = _return_no_host
        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        extra_spec = {}
        self.assertRaises(exception.NoValidHost,
                          self.driver_cls().select_destinations, self.context,
                          containers, extra_spec)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as n_const
from neutronclient.common import exceptions as n_exceptions
from neutronclient.neutron import v2_0 as neutronv20
from oslo_log import log as logging
from oslo_utils import uuidutils

from zun.common import clients
from zun.common import context as zun_context
from zun.common import exception
from zun.common.i18n import _


LOG = logging.getLogger(__name__)


class NeutronAPI(object):

    def __init__(self, context):
        self.context = context
        self.client = clients.OpenStackClients(self.context).neutron()
        self.admin_client = None

    def __getattr__(self, key):
        return getattr(self.client, key)

    def _get_admin_client(self):
        if self.admin_client is None:
            context = zun_context.get_admin_context()
            self.admin_client = clients.OpenStackClients(context).neutron()
        return self.admin_client

    def update_port(self, port, body=None, admin=False):
        if admin:
            client = self._get_admin_client()
        else:
            client = self.client
        return client.update_port(port, body)

    def create_port(self, body=None, admin=False):
        if admin:
            client = self._get_admin_client()
        else:
            client = self.client
        return client.create_port(body)

    def find_resourceid_by_name_or_id(self, resource, name_or_id,
                                      project_id=None):
        return neutronv20.find_resourceid_by_name_or_id(
            self.client, resource, name_or_id, project_id)

    def get_available_network(self):
        search_opts = {'tenant_id': self.context.project_id, 'shared': False}
        # NOTE(kiennt): Pick shared network if no tenant network
        nets = self.list_networks(**search_opts).get('networks', []) or \
            self.list_networks(**{'shared': True}).get('networks', [])
        if not nets:
            raise exception.Conflict(_(
                "There is no neutron network available"))
        nets.sort(key=lambda x: x['created_at'])
        return nets[0]

    def get_neutron_network(self, network):
        if uuidutils.is_uuid_like(network):
            networks = self.list_networks(id=network)['networks']
        else:
            networks = self.list_networks(name=network)['networks']

        if len(networks) == 0:
            raise exception.NetworkNotFound(network=network)
        elif len(networks) > 1:
            raise exception.Conflict(_(
                'Multiple neutron networks exist with same name. '
                'Please use the uuid instead.'))

        network = networks[0]
        return network

    def get_neutron_port(self, port):
        if uuidutils.is_uuid_like(port):
            ports = self.list_ports(id=port)['ports']
        else:
            ports = self.list_ports(name=port)['ports']

        if len(ports) == 0:
            raise exception.PortNotFound(port=port)
        elif len(ports) > 1:
            raise exception.Conflict(_(
                'Multiple neutron ports exist with same name. '
                'Please use the uuid instead.'))

        port = ports[0]
        return port

    def ensure_neutron_port_usable(self, port, project_id=None):
        if project_id is None:
            project_id = self.context.project_id

        # Make sure the container has access to the port.
        if port['tenant_id'] != project_id:
            raise exception.PortNotUsable(port=port['id'])

        # Make sure the port isn't already attached to another
        # container or Nova instance.
        if port.get('device_id'):
            raise exception.PortInUse(port=port['id'])

        if port.get('status') in (n_const.PORT_STATUS_ACTIVE,
                                  n_const.PORT_STATUS_BUILD,
                                  n_const.PORT_STATUS_ERROR):
            raise exception.PortNotUsable(port=port['id'])

        # Make sure the port is usable
        binding_vif_type = port.get('binding:vif_type')
        if binding_vif_type == 'binding_failed':
            raise exception.PortBindingFailed(port=port['id'])

    def expose_ports(self, secgroup_id, ports):
        for port in ports:
            port, proto = port.split('/')
            secgroup_rule = {
                'security_group_id': secgroup_id,
                'direction': 'ingress',
                'port_range_min': port,
                'port_range_max': port,
                'protocol': proto,
                'remote_ip_prefix': '0.0.0.0/0',
            }

            try:
                self.create_security_group_rule({
                    'security_group_rule': secgroup_rule})
            except n_exceptions.NeutronClientException as ex:
                LOG.error("Error happened during creating a "
                          "Neutron security group "
                          "rule: %s", ex)
                self.delete_security_group(secgroup_id)
                raise exception.ZunException(_(
                    "Could not create required security group rules %s "
                    "for setting up exported port.") % secgroup_rule)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
import six

from zun.common import exception
import zun.conf
from zun.db import api as dbapi
from zun.tests.unit.db import base
from zun.tests.unit.db import utils


CONF = zun.conf.CONF


class DbRegistryTestCase(base.DbTestCase):

    def setUp(self):
        super(DbRegistryTestCase, self).setUp()

    def test_create_registry(self):
        username = 'fake-user'
        password = 'fake-pass'
        registry = utils.create_test_registry(context=self.context,
                                              username=username,
                                              password=password)
        self.assertEqual(username, registry.username)
        self.assertEqual(password, registry.password)

    def test_create_registry_already_exists(self):
        utils.create_test_registry(context=self.context,
                                   uuid='123')
        with self.assertRaisesRegex(exception.RegistryAlreadyExists,
                                    'A registry with UUID 123.*'):
            utils.create_test_registry(context=self.context,
                                       uuid='123')

    def test_get_registry_by_uuid(self):
        username = 'fake-user'
        password = 'fake-pass'
        registry = utils.create_test_registry(context=self.context,
                                              username=username,
                                              password=password)
        res = dbapi.get_registry_by_uuid(self.context,
                                         registry.uuid)
        self.assertEqual(registry.id, res.id)
        self.assertEqual(registry.uuid, res.uuid)
        self.assertEqual(username, res.username)
        self.assertEqual(password, res.password)

    def test_get_registry_by_name(self):
        username = 'fake-user'
        password = 'fake-pass'
        registry = utils.create_test_registry(context=self.context,
                                              username=username,
                                              password=password)
        res = dbapi.get_registry_by_name(
            self.context, registry.name)
        self.assertEqual(registry.id, res.id)
        self.assertEqual(registry.uuid, res.uuid)
        self.assertEqual(username, res.username)
        self.assertEqual(password, res.password)

    def test_get_registry_that_does_not_exist(self):
        self.assertRaises(exception.RegistryNotFound,
                          dbapi.get_registry_by_uuid,
                          self.context,
                          uuidutils.generate_uuid())

    def test_list_registries(self):
        uuids = []
        passwords = []
        for i in range(1, 6):
            password = 'pass' + str(i)
            passwords.append(password)
            registry = utils.create_test_registry(
                uuid=uuidutils.generate_uuid(),
                context=self.context,
                name='registry' + str(i),
                password=password)
            uuids.append(six.text_type(registry['uuid']))
        res = dbapi.list_registries(self.context)
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), sorted(res_uuids))
        res_passwords = [r.password for r in res]
        self.assertEqual(sorted(passwords), sorted(res_passwords))

    def test_list_registries_sorted(self):
        uuids = []
        for i in range(5):
            registry = utils.create_test_registry(
                uuid=uuidutils.generate_uuid(),
                context=self.context,
                name='registry' + str(i))
            uuids.append(six.text_type(registry.uuid))
        res = dbapi.list_registries(self.context, sort_key='uuid')
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), res_uuids)

        self.assertRaises(exception.InvalidParameterValue,
                          dbapi.list_registries,
                          self.context,
                          sort_key='foo')

    def test_list_registries_with_filters(self):
        registry1 = utils.create_test_registry(
            name='registry-one',
            uuid=uuidutils.generate_uuid(),
            context=self.context)
        registry2 = utils.create_test_registry(
            name='registry-two',
            uuid=uuidutils.generate_uuid(),
            context=self.context)

        res = dbapi.list_registries(
            self.context, filters={'name': 'registry-one'})
        self.assertEqual([registry1.id], [r.id for r in res])

        res = dbapi.list_registries(
            self.context, filters={'name': 'registry-two'})
        self.assertEqual([registry2.id], [r.id for r in res])

        res = dbapi.list_registries(
            self.context, filters={'name': 'bad-registry'})
        self.assertEqual([], [r.id for r in res])

        res = dbapi.list_registries(
            self.context,
            filters={'name': registry1.name})
        self.assertEqual([registry1.id], [r.id for r in res])

    def test_list_registries_with_list_filters(self):
        registry1 = utils.create_test_registry(
            name='registry-one',
            uuid=uuidutils.generate_uuid(),
            context=self.context)
        registry2 = utils.create_test_registry(
            name='registry-two',
            uuid=uuidutils.generate_uuid(),
            context=self.context)

        res = dbapi.list_registries(
            self.context, filters={'name': ['registry-one', 'registry-two']})
        uuids = sorted([registry1.uuid, registry2.uuid])
        self.assertEqual(uuids, sorted([r.uuid for r in res]))

    def test_destroy_registry(self):
        registry = utils.create_test_registry(context=self.context)
        dbapi.destroy_registry(self.context, registry.id)
        self.assertRaises(exception.RegistryNotFound,
                          dbapi.get_registry_by_uuid,
                          self.context, registry.uuid)

    def test_destroy_registry_by_uuid(self):
        registry = utils.create_test_registry(context=self.context)
        dbapi.destroy_registry(self.context, registry.uuid)
        self.assertRaises(exception.RegistryNotFound,
                          dbapi.get_registry_by_uuid,
                          self.context, registry.uuid)

    def test_destroy_registry_that_does_not_exist(self):
        self.assertRaises(exception.RegistryNotFound,
                          dbapi.destroy_registry, self.context,
                          uuidutils.generate_uuid())

    def test_update_registry(self):
        registry = utils.create_test_registry(context=self.context)
        old_name = registry.name
        new_name = 'new-name'
        new_password = 'new-pass'
        self.assertNotEqual(old_name, new_name)

        res = dbapi.update_registry(self.context, registry.id,
                                    {'name': new_name,
                                     'password': new_password})
        self.assertEqual(new_name, res.name)
        self.assertEqual(new_password, res.password)

    def test_update_registry_not_found(self):
        registry_uuid = uuidutils.generate_uuid()
        new_name = 'new-name'
        self.assertRaises(exception.RegistryNotFound,
                          dbapi.update_registry, self.context,
                          registry_uuid, {'name': new_name})

    def test_update_registry_uuid(self):
        registry = utils.create_test_registry(context=self.context)
        self.assertRaises(exception.InvalidParameterValue,
                          dbapi.update_registry, self.context,
                          registry.id, {'uuid': ''})



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Zun Service Layer"""

from oslo_log import log
from oslo_service import periodic_task

from zun.common import context
from zun import objects


LOG = log.getLogger(__name__)


class ZunServicePeriodicTasks(periodic_task.PeriodicTasks):
    """Zun periodic Task class

    Any periodic task job need to be added into this class
    """

    def __init__(self, conf, binary):
        self.zun_service_ref = None
        self.host = conf.host
        self.binary = binary
        self.availability_zone = conf.default_availability_zone
        super(ZunServicePeriodicTasks, self).__init__(conf)

    @periodic_task.periodic_task(run_immediately=True)
    @context.set_context
    def update_zun_service(self, ctx):
        LOG.debug('Update zun_service')
        if self.zun_service_ref is None:
            self.zun_service_ref = \
                objects.ZunService.get_by_host_and_binary(
                    ctx, self.host, self.binary)
            if self.zun_service_ref is None:
                zun_service_dict = {
                    'host': self.host,
                    'binary': self.binary,
                    'availability_zone': self.availability_zone
                }
                self.zun_service_ref = objects.ZunService(
                    ctx, **zun_service_dict)
                self.zun_service_ref.create()
            else:
                self.zun_service_ref.availability_zone = \
                    self.availability_zone
        self.zun_service_ref.report_state_up()


def setup(conf, binary, tg):
    pt = ZunServicePeriodicTasks(conf, binary)
    tg.add_dynamic_timer(
        pt.run_periodic_tasks,
        periodic_interval_max=conf.periodic_interval_max,
        context=None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Zun DB test base class."""

import fixtures

import zun.conf
from zun.db import api as db_api
from zun.db.sqlalchemy import api as sqla_api
from zun.db.sqlalchemy import migration
from zun.db.sqlalchemy import models
from zun.tests import base


CONF = zun.conf.CONF

_DB_CACHE = None


class Database(fixtures.Fixture):

    def __init__(self, db_api, db_migrate, sql_connection):
        self.sql_connection = sql_connection
        self.engine = db_api.get_engine()
        self.engine.dispose()
        conn = self.engine.connect()
        self.setup_sqlite(db_migrate)
        self.post_migrations()

        self._DB = "".join(line for line in conn.connection.iterdump())
        self.engine.dispose()

    def setup_sqlite(self, db_migrate):
        if db_migrate.version():
            return
        models.Base.metadata.create_all(self.engine)
        db_migrate.stamp('head')

    def _setUp(self):
        conn = self.engine.connect()
        conn.connection.executescript(self._DB)
        self.addCleanup(self.engine.dispose)

    def post_migrations(self):
        """Any addition steps that are needed outside of the migrations."""


class DbTestCase(base.TestCase):

    def setUp(self):
        super(DbTestCase, self).setUp()

        self.dbapi = db_api._get_dbdriver_instance()

        global _DB_CACHE
        if not _DB_CACHE:
            _DB_CACHE = Database(sqla_api, migration,
                                 sql_connection=CONF.database.connection)
        self.useFixture(_DB_CACHE)


class ModelsObjectComparatorMixin(object):
    def _dict_from_object(self, obj, ignored_keys):
        if ignored_keys is None:
            ignored_keys = []

        return {k: v for k, v in obj.items()
                if k not in ignored_keys}

    def _assertEqualObjects(self, obj1, obj2, ignored_keys=None):
        obj1 = self._dict_from_object(obj1, ignored_keys)
        obj2 = self._dict_from_object(obj2, ignored_keys)

        self.assertEqual(len(obj1),
                         len(obj2),
                         "Keys mismatch: %s" %
                         str(set(obj1.keys()) ^ set(obj2.keys())))
        for key, value in obj1.items():
            self.assertEqual(value, obj2[key])

    def _assertEqualListsOfObjects(self, objs1, objs2, ignored_keys=None):
        obj_to_dict = lambda o: self._dict_from_object(o, ignored_keys)
        sort_key = lambda d: [d[k] for k in sorted(d)]
        conv_and_sort = lambda obj: sorted(map(obj_to_dict, obj), key=sort_key)

        self.assertEqual(conv_and_sort(objs1), conv_and_sort(objs2))

    def _assertEqualOrderedListOfObjects(self, objs1, objs2,
                                         ignored_keys=None):
        obj_to_dict = lambda o: self._dict_from_object(o, ignored_keys)
        conv = lambda objs: [obj_to_dict(obj) for obj in objs]

        self.assertEqual(conv(objs1), conv(objs2))

    def _assertEqualListsOfPrimitivesAsSets(self, primitives1, primitives2):
        self.assertEqual(len(primitives1), len(primitives2))
        for primitive in primitives1:
            self.assertIn(primitive, primitives2)

        for primitive in primitives2:
            self.assertIn(primitive, primitives1)



# Copyright (c) 2013 dotCloud, Inc.
# Copyright 2014 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import inspect

from oslo_config import cfg
import six

import docker

CONF = cfg.CONF
DEFAULT_TIMEOUT_SECONDS = 120
DEFAULT_DOCKER_API_VERSION = '1.21'


def filter_data(f):
    """Decorator that post-processes data returned by Docker.

     This will avoid any surprises with different versions of Docker.
    """
    @functools.wraps(f, assigned=[])
    def wrapper(*args, **kwds):
        out = f(*args, **kwds)

        def _filter(obj):
            if isinstance(obj, list):
                new_list = []
                for o in obj:
                    new_list.append(_filter(o))
                obj = new_list
            if isinstance(obj, dict):
                for k, v in obj.items():
                    if isinstance(k, six.string_types):
                        obj[k.lower()] = _filter(v)
            return obj
        return _filter(out)
    return wrapper


class DockerHTTPClient(docker.APIClient):
    def __init__(self, url='unix://var/run/docker.sock'):
        if (CONF.docker.cert_file or
                CONF.docker.key_file):
            client_cert = (CONF.docker.cert_file, CONF.docker.key_file)
        else:
            client_cert = None
        if (CONF.docker.ca_file or
                CONF.docker.api_insecure or
                client_cert):
            ssl_config = docker.tls.TLSConfig(
                client_cert=client_cert,
                ca_cert=CONF.docker.ca_file,
                verify=CONF.docker.api_insecure)
        else:
            ssl_config = False
        super(DockerHTTPClient, self).__init__(
            base_url=url,
            version=DEFAULT_DOCKER_API_VERSION,
            timeout=DEFAULT_TIMEOUT_SECONDS,
            tls=ssl_config
        )
        self._setup_decorators()

    def _setup_decorators(self):
        # NOTE(junbo.li): we need to distinguish class methods types
        # for py2 and py3, because the concept of 'unbound methods' has
        # been removed from the python3.x
        if six.PY3:
            member_type = inspect.isfunction
        else:
            member_type = inspect.ismethod
        for name, member in inspect.getmembers(self, member_type):
            if not name.startswith('_'):
                setattr(self, name, filter_data(member))

    def pause(self, container_id):
        url = self._url("/containers/{0}/pause".format(container_id))
        res = self._post(url)
        return res.status_code == 204

    def unpause(self, container_id):
        url = self._url("/containers/{0}/unpause".format(container_id))
        res = self._post(url)
        return res.status_code == 204

    def load_repository_file(self, name, path):
        with open(path, 'rb') as fh:
            self.load_image(fh)

    def get_container_logs(self, container_id):
        return self.attach(container_id, 1, 1, 0, 1)



# Copyright 2016 Intel.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from docker import errors
import six

from oslo_log import log as logging
from oslo_utils import excutils

from zun.common.docker_image import reference as docker_image
from zun.common import exception
from zun.common.i18n import _
from zun.common import utils
import zun.conf
from zun.container.docker import utils as docker_utils
from zun.image import driver


CONF = zun.conf.CONF

LOG = logging.getLogger(__name__)


class DockerDriver(driver.ContainerImageDriver):
    def __init__(self):
        super(DockerDriver, self).__init__()

    def delete_image(self, context, img_id):
        LOG.debug('Delete an image %s in docker', img_id)
        with docker_utils.docker_client() as docker:
            try:
                docker.remove_image(img_id)
            except errors.ImageNotFound:
                return
            except errors.APIError as api_error:
                raise exception.ZunException(six.text_type(api_error))
            except Exception as e:
                LOG.exception('Unknown exception occurred while deleting '
                              'image %s in glance:%s',
                              img_id,
                              six.text_type(e))
                raise exception.ZunException(six.text_type(e))

    def _search_image_on_host(self, repo, tag):
        with docker_utils.docker_client() as docker:
            image = repo + ":" + tag
            LOG.debug('Inspecting image locally %s', image)
            try:
                image_dict = docker.inspect_image(image)
                if image_dict:
                    return {'image': repo, 'path': None}
            except errors.NotFound:
                LOG.debug('Image %s not found locally', image)
                return None

    def _pull_image(self, repo, tag, registry):
        auth_config = None
        image_ref = docker_image.Reference.parse(repo)
        registry_domain, remainder = image_ref.split_hostname()
        if registry and registry.username:
            auth_config = {'username': registry.username,
                           'password': registry.password}
        elif (registry_domain and
                registry_domain == CONF.docker.default_registry and
                CONF.docker.default_registry_username):
            auth_config = {'username': CONF.docker.default_registry_username,
                           'password': CONF.docker.default_registry_password}

        with docker_utils.docker_client() as docker:
            try:
                docker.pull(repo, tag=tag, auth_config=auth_config)
            except errors.NotFound as e:
                raise exception.ImageNotFound(message=six.text_type(e))
            except errors.APIError as e:
                LOG.exception('Error on pulling image')
                message = _('Error on pulling image: %(repo)s:%(tag)s') % {
                    'repo': repo, 'tag': tag}
                raise exception.ZunException(message)

    def pull_image(self, context, repo, tag, image_pull_policy, registry):
        image_loaded = True
        image = self._search_image_on_host(repo, tag)
        if not utils.should_pull_image(image_pull_policy, bool(image)):
            if image:
                LOG.debug('Image  %s present locally', repo)
                return image, image_loaded
            else:
                message = _('Image %s not present with pull policy of Never'
                            ) % repo
                raise exception.ImageNotFound(message)

        try:
            LOG.debug('Pulling image from docker %(repo)s,'
                      ' context %(context)s',
                      {'repo': repo, 'context': context})
            self._pull_image(repo, tag, registry)
            return {'image': repo, 'path': None}, image_loaded
        except exception.ImageNotFound:
            with excutils.save_and_reraise_exception():
                LOG.error('Image %s was not found in docker repo', repo)
        except exception.DockerError:
            with excutils.save_and_reraise_exception():
                LOG.error('Docker API error occurred during downloading '
                          'image %s', repo)
        except Exception as e:
            msg = _('Cannot download image from docker: {0}')
            raise exception.ZunException(msg.format(e))

    def search_image(self, context, repo, tag, exact_match):
        image_ref = docker_image.Reference.parse(repo)
        registry, image_name = image_ref.split_hostname()
        if registry and registry != 'docker.io':
            # Images searching is only supported in DockerHub
            msg = _('Image searching is not supported in registry: {0}')
            raise exception.OperationNotSupported(msg.format(registry))

        with docker_utils.docker_client() as docker:
            try:
                # TODO(hongbin): search image by both name and tag
                images = docker.search(image_name)
            except errors.APIError as api_error:
                raise exception.ZunException(six.text_type(api_error))
            except Exception as e:
                msg = _('Cannot search image in docker: {0}')
                raise exception.ZunException(msg.format(e))

        if exact_match:
            images = [i for i in images if i['name'] == image_name]

        for image in images:
            image['metadata'] = {}
            for key in ('is_official', 'star_count'):
                value = image.pop(key, None)
                if value is not None:
                    image['metadata'][key] = value

        # TODO(hongbin): convert images to a list of Zun Image object
        return images



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
from oslo_utils import versionutils
from oslo_versionedobjects import fields

from zun.objects import base


@base.ZunObjectRegistry.register
class ContainerPCIRequest(base.ZunPersistentObject, base.ZunObject):
    # Version 1.0: Add request_id
    VERSION = '1.0'

    fields = {
        'count': fields.IntegerField(),
        'spec': fields.ListOfDictOfNullableStringsField(),
        'alias_name': fields.StringField(nullable=True),
        # Note(moshele): is_new is deprecated and should be removed
        # on major version bump
        'is_new': fields.BooleanField(default=False),
        'request_id': fields.UUIDField(nullable=True),
    }

    def obj_load_attr(self, attr):
        setattr(self, attr, None)

    def obj_make_compatible(self, primitive, target_version):
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1) and 'request_id' in primitive:
            del primitive['request_id']


@base.ZunObjectRegistry.register
class ContainerPCIRequests(base.ZunPersistentObject, base.ZunObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'container_uuid': fields.UUIDField(),
        'requests': fields.ListOfObjectsField('ContainerPCIRequest'),
    }

    def to_json(self):
        blob = [{'count': x.count,
                 'spec': x.spec,
                 'alias_name': x.alias_name,
                 'is_new': x.is_new,
                 'request_id': x.request_id} for x in self.requests]
        return jsonutils.dumps(blob)

    @classmethod
    def from_request_spec_container_props(cls, pci_requests):
        objs = [ContainerPCIRequest(**request)
                for request in pci_requests['requests']]
        return cls(requests=objs,
                   container_uuid=pci_requests['container_uuid'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import fields

from zun.db import api as dbapi
from zun.objects import base


LOG = logging.getLogger(__name__)


@base.ZunObjectRegistry.register
class Registry(base.ZunPersistentObject, base.ZunObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.IntegerField(),
        'uuid': fields.UUIDField(nullable=True),
        'name': fields.StringField(nullable=True),
        'project_id': fields.StringField(nullable=True),
        'user_id': fields.StringField(nullable=True),
        'domain': fields.StringField(nullable=True),
        'username': fields.StringField(nullable=True),
        'password': fields.StringField(nullable=True),
    }

    @staticmethod
    def _from_db_object(registry, db_registry):
        """Converts a database entity to a formal object."""
        for field in registry.fields:
            setattr(registry, field, db_registry[field])

        registry.obj_reset_changes()
        return registry

    @staticmethod
    def _from_db_object_list(db_objects, cls, context):
        """Converts a list of database entities to a list of formal objects."""
        return [Registry._from_db_object(cls(context), obj)
                for obj in db_objects]

    @base.remotable_classmethod
    def get_by_id(cls, context, id):
        """Find a registry based on id and return a :class:`Registry` object.

        :param id: the id of a registry.
        :param context: Security context
        :returns: a :class:`Registry` object.
        """
        db_registry = dbapi.get_registry_by_id(context, id)
        registry = Registry._from_db_object(cls(context), db_registry)
        return registry

    @base.remotable_classmethod
    def get_by_uuid(cls, context, uuid):
        """Find a registry based on uuid and return a :class:`Registry` object.

        :param uuid: the uuid of a registry.
        :param context: Security context
        :returns: a :class:`Registry` object.
        """
        db_registry = dbapi.get_registry_by_uuid(context, uuid)
        registry = Registry._from_db_object(cls(context), db_registry)
        return registry

    @base.remotable_classmethod
    def get_by_name(cls, context, name):
        """Find a registry based on name and return a Registry object.

        :param name: the logical name of a registry.
        :param context: Security context
        :returns: a :class:`Registry` object.
        """
        db_registry = dbapi.get_registry_by_name(context, name)
        registry = Registry._from_db_object(cls(context), db_registry)
        return registry

    @base.remotable_classmethod
    def list(cls, context, limit=None, marker=None,
             sort_key=None, sort_dir=None, filters=None):
        """Return a list of Registry objects.

        :param context: Security context.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :param filters: filters when list registries.
        :returns: a list of :class:`Registry` object.

        """
        db_registries = dbapi.list_registries(
            context, limit=limit, marker=marker, sort_key=sort_key,
            sort_dir=sort_dir, filters=filters)
        return Registry._from_db_object_list(db_registries, cls, context)

    @base.remotable
    def create(self, context):
        """Create a Registry record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Registry(context)

        """
        values = self.obj_get_changes()
        db_registry = dbapi.create_registry(context, values)
        self._from_db_object(self, db_registry)

    @base.remotable
    def destroy(self, context=None):
        """Delete the Registry from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Registry(context)
        """
        dbapi.destroy_registry(context, self.uuid)
        self.obj_reset_changes()

    @base.remotable
    def save(self, context=None):
        """Save updates to this Registry.

        Updates will be made column by column based on the result
        of self.what_changed().

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Registry(context)
        """
        updates = self.obj_get_changes()
        dbapi.update_registry(context, self.uuid, updates)

        self.obj_reset_changes()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

import zun.conf
from zun import objects
from zun.servicegroup import zun_service_periodic as periodic
from zun.tests import base


class ZunServicePeriodicTestCase(base.BaseTestCase):
    def setUp(self):
        super(ZunServicePeriodicTestCase, self).setUp()
        mock_zun_service_refresh = mock.Mock()

        class FakeSrv(object):
            report_state_up = mock_zun_service_refresh

        self.fake_srv = FakeSrv()
        self.fake_srv_refresh = mock_zun_service_refresh

    @mock.patch.object(objects.ZunService, 'get_by_host_and_binary')
    @mock.patch.object(objects.ZunService, 'create')
    @mock.patch.object(objects.ZunService, 'report_state_up')
    def test_update_zun_service_firsttime(self,
                                          mock_srv_refresh,
                                          mock_srv_create,
                                          mock_srv_get
                                          ):
        p_task = periodic.ZunServicePeriodicTasks(zun.conf.CONF,
                                                  'fake-conductor')
        mock_srv_get.return_value = None

        p_task.update_zun_service(None)

        mock_srv_get.assert_called_once_with(mock.ANY, p_task.host,
                                             p_task.binary)
        mock_srv_create.assert_called_once_with()
        mock_srv_refresh.assert_called_once_with()

    @mock.patch.object(objects.ZunService, 'get_by_host_and_binary')
    @mock.patch.object(objects.ZunService, 'create')
    def test_update_zun_service_on_restart(self,
                                           mock_srv_create,
                                           mock_srv_get):
        p_task = periodic.ZunServicePeriodicTasks(zun.conf.CONF,
                                                  'fake-conductor')
        mock_srv_get.return_value = self.fake_srv

        p_task.update_zun_service(None)

        mock_srv_get.assert_called_once_with(mock.ANY, p_task.host,
                                             p_task.binary)
        self.fake_srv_refresh.assert_called_once_with()

    def test_update_zun_service_regular(self):
        p_task = periodic.ZunServicePeriodicTasks(zun.conf.CONF,
                                                  'fake-conductor')
        p_task.zun_service_ref = self.fake_srv

        p_task.update_zun_service(None)

        self.fake_srv_refresh.assert_called_once_with()
